# Copyright (C) 2021  Nexedi SA and Contributors.
#                     Romain Courteaud <romain@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

import unittest
from surykatka.db import LogDB
import peewee
import datetime
from surykatka.domain import logWhoisQuery, packDomain, queryWhois
from surykatka.status import logStatus
import mock


class MockAnswer(object):
    def __init__(
        self,
        registrar,
        whois_server,
        creation_date,
        updated_date,
        expiration_date,
        name_servers,
        status,
        emails,
        dnssec,
        name,
        org,
        address,
        city,
        state,
        zipcode,
        country,
    ):
        self.registrar = registrar
        self.whois_server = whois_server
        self.creation_date = creation_date
        self.updated_date = updated_date
        self.expiration_date = expiration_date
        self.name_servers = name_servers
        self.status = status
        self.emails = emails
        self.dnssec = dnssec
        self.name = name
        self.org = org
        self.address = address
        self.city = city
        self.state = state
        self.zipcode = zipcode
        self.country = country


class SurykatkaDomainTestCase(unittest.TestCase):
    def setUp(self):
        self.db = LogDB(":memory:")
        self.db.createTables()

    ################################################
    # logWhoisQuery
    ################################################
    def test_logWhoisQuery_insertFirst(self):
        domain = "http://example.org"
        registrar = "REGISTRAR"
        whois_server = "WHOISSERVER"
        creation_date = datetime.datetime.utcnow()
        updated_date = datetime.datetime.utcnow()
        expiration_date = datetime.datetime.utcnow()
        name_servers = "NAMESERVER"
        whois_status = "WHOISSTATUS"
        emails = "EMAIL"
        dnssec = "DNSSEC"
        name = "NAME"
        org = "ORG"
        address = "ADDRESS"
        city = "CITY"
        state = "STATE"
        zipcode = "ZIPCODE"
        country = "COUNTRY"
        status_id = logStatus(self.db, "foo")
        result = logWhoisQuery(
            self.db,
            status_id,
            domain,
            registrar,
            whois_server,
            creation_date,
            updated_date,
            expiration_date,
            name_servers,
            whois_status,
            emails,
            dnssec,
            name,
            org,
            address,
            city,
            state,
            zipcode,
            country,
        )
        assert self.db.DomainChange.select().count() == 1
        assert self.db.DomainChange.get().domain == domain
        assert self.db.DomainChange.get().registrar == registrar
        assert self.db.DomainChange.get().whois_server == whois_server
        assert self.db.DomainChange.get().creation_date == creation_date
        assert self.db.DomainChange.get().updated_date == updated_date
        assert self.db.DomainChange.get().expiration_date == expiration_date
        assert self.db.DomainChange.get().name_servers == name_servers
        assert self.db.DomainChange.get().whois_status == whois_status
        assert self.db.DomainChange.get().emails == emails
        assert self.db.DomainChange.get().dnssec == dnssec
        assert self.db.DomainChange.get().name == name
        assert self.db.DomainChange.get().org == org
        assert self.db.DomainChange.get().address == address
        assert self.db.DomainChange.get().city == city
        assert self.db.DomainChange.get().state == state
        assert self.db.DomainChange.get().zipcode == zipcode
        assert self.db.DomainChange.get().country == country
        assert self.db.DomainChange.get().status_id == status_id
        assert result == status_id

    def test_logWhoisQuery_insertOnlyOnePerStatusIdDomain(self):
        domain = "http://example.org"
        registrar = "REGISTRAR"
        whois_server = "WHOISSERVER"
        creation_date = datetime.datetime.utcnow()
        updated_date = datetime.datetime.utcnow()
        expiration_date = datetime.datetime.utcnow()
        name_servers = "NAMESERVER"
        whois_status = "WHOISSTATUS"
        emails = "EMAIL"
        dnssec = "DNSSEC"
        name = "NAME"
        org = "ORG"
        address = "ADDRESS"
        city = "CITY"
        state = "STATE"
        zipcode = "ZIPCODE"
        country = "COUNTRY"
        status_id = logStatus(self.db, "foo")
        logWhoisQuery(
            self.db,
            status_id,
            domain,
            registrar,
            whois_server,
            creation_date,
            updated_date,
            expiration_date,
            name_servers,
            whois_status,
            emails,
            dnssec,
            name,
            org,
            address,
            city,
            state,
            zipcode,
            country,
        )
        country += "FOO"
        try:
            logWhoisQuery(
                self.db,
                status_id,
                domain,
                registrar,
                whois_server,
                creation_date,
                updated_date,
                expiration_date,
                name_servers,
                whois_status,
                emails,
                dnssec,
                name,
                org,
                address,
                city,
                state,
                zipcode,
                country,
            )
        except peewee.IntegrityError:
            assert self.db.DomainChange.select().count() == 1
            assert self.db.DomainChange.get().status_id == status_id
        else:
            raise NotImplementedError("Expected IntegrityError")

    def test_logDomainQuery_skipIdenticalPreviousValues(self):
        domain = "http://example.org"
        registrar = "REGISTRAR"
        whois_server = "WHOISSERVER"
        creation_date = datetime.datetime.utcnow()
        updated_date = datetime.datetime.utcnow()
        expiration_date = datetime.datetime.utcnow()
        name_servers = "NAMESERVER"
        whois_status = "WHOISSTATUS"
        emails = "EMAIL"
        dnssec = "DNSSEC"
        name = "NAME"
        org = "ORG"
        address = "ADDRESS"
        city = "CITY"
        state = "STATE"
        zipcode = "ZIPCODE"
        country = "COUNTRY"
        status_id = logStatus(self.db, "foo")
        result = logWhoisQuery(
            self.db,
            status_id,
            domain,
            registrar,
            whois_server,
            creation_date,
            updated_date,
            expiration_date,
            name_servers,
            whois_status,
            emails,
            dnssec,
            name,
            org,
            address,
            city,
            state,
            zipcode,
            country,
        )

        status_id_2 = logStatus(self.db, "foo")
        result_2 = logWhoisQuery(
            self.db,
            status_id_2,
            domain,
            registrar,
            whois_server,
            creation_date,
            updated_date,
            expiration_date,
            name_servers,
            whois_status,
            emails,
            dnssec,
            name,
            org,
            address,
            city,
            state,
            zipcode,
            country,
        )

        assert self.db.DomainChange.select().count() == 1
        assert self.db.DomainChange.get().domain == domain
        assert self.db.DomainChange.get().registrar == registrar
        assert self.db.DomainChange.get().whois_server == whois_server
        assert self.db.DomainChange.get().creation_date == creation_date
        assert self.db.DomainChange.get().updated_date == updated_date
        assert self.db.DomainChange.get().expiration_date == expiration_date
        assert self.db.DomainChange.get().name_servers == name_servers
        assert self.db.DomainChange.get().whois_status == whois_status
        assert self.db.DomainChange.get().emails == emails
        assert self.db.DomainChange.get().dnssec == dnssec
        assert self.db.DomainChange.get().name == name
        assert self.db.DomainChange.get().org == org
        assert self.db.DomainChange.get().address == address
        assert self.db.DomainChange.get().city == city
        assert self.db.DomainChange.get().state == state
        assert self.db.DomainChange.get().zipcode == zipcode
        assert self.db.DomainChange.get().country == country
        assert self.db.DomainChange.get().status_id == status_id
        assert result == status_id
        assert result_2 == status_id

    def test_logDomainQuery_insertWhenDifferentAnswer(self):
        domain = "http://example.org"
        registrar = "REGISTRAR"
        registrar_2 = registrar + "foo"
        whois_server = "WHOISSERVER"
        whois_server_2 = whois_server + "foo"
        creation_date = datetime.datetime(2000, 1, 1)
        creation_date_2 = datetime.datetime(2000, 1, 2)
        updated_date = datetime.datetime(2000, 1, 3)
        updated_date_2 = datetime.datetime(2000, 1, 4)
        expiration_date = datetime.datetime(2000, 1, 5)
        expiration_date_2 = datetime.datetime(2000, 1, 6)
        name_servers = "NAMESERVER"
        name_servers_2 = name_servers + "foo"
        whois_status = "WHOISSTATUS"
        whois_status_2 = whois_status + "foo"
        emails = "EMAIL"
        emails_2 = emails + "foo"
        dnssec = "DNSSEC"
        dnssec_2 = dnssec + "foo"
        name = "NAME"
        name_2 = name + "foo"
        org = "ORG"
        org_2 = org + "foo"
        address = "ADDRESS"
        address_2 = address + "foo"
        city = "CITY"
        city_2 = city + "foo"
        state = "STATE"
        state_2 = state + "foo"
        zipcode = "ZIPCODE"
        zipcode_2 = zipcode + "foo"
        country = "COUNTRY"
        country_2 = country + "foo"

        logWhoisQuery(
            self.db,
            logStatus(self.db, "foo"),
            domain,
            registrar,
            whois_server,
            creation_date,
            updated_date,
            expiration_date,
            name_servers,
            whois_status,
            emails,
            dnssec,
            name,
            org,
            address,
            city,
            state,
            zipcode,
            country,
        )
        assert self.db.DomainChange.select().count() == 1

        logWhoisQuery(
            self.db,
            logStatus(self.db, "foo"),
            domain,
            registrar_2,
            whois_server,
            creation_date,
            updated_date,
            expiration_date,
            name_servers,
            whois_status,
            emails,
            dnssec,
            name,
            org,
            address,
            city,
            state,
            zipcode,
            country,
        )
        assert self.db.DomainChange.select().count() == 2

        logWhoisQuery(
            self.db,
            logStatus(self.db, "foo"),
            domain,
            registrar_2,
            whois_server_2,
            creation_date,
            updated_date,
            expiration_date,
            name_servers,
            whois_status,
            emails,
            dnssec,
            name,
            org,
            address,
            city,
            state,
            zipcode,
            country,
        )
        assert self.db.DomainChange.select().count() == 3

        logWhoisQuery(
            self.db,
            logStatus(self.db, "foo"),
            domain,
            registrar_2,
            whois_server_2,
            creation_date_2,
            updated_date,
            expiration_date,
            name_servers,
            whois_status,
            emails,
            dnssec,
            name,
            org,
            address,
            city,
            state,
            zipcode,
            country,
        )
        assert self.db.DomainChange.select().count() == 4

        logWhoisQuery(
            self.db,
            logStatus(self.db, "foo"),
            domain,
            registrar_2,
            whois_server_2,
            creation_date_2,
            updated_date_2,
            expiration_date,
            name_servers,
            whois_status,
            emails,
            dnssec,
            name,
            org,
            address,
            city,
            state,
            zipcode,
            country,
        )
        assert self.db.DomainChange.select().count() == 5

        logWhoisQuery(
            self.db,
            logStatus(self.db, "foo"),
            domain,
            registrar_2,
            whois_server_2,
            creation_date_2,
            updated_date_2,
            expiration_date_2,
            name_servers,
            whois_status,
            emails,
            dnssec,
            name,
            org,
            address,
            city,
            state,
            zipcode,
            country,
        )
        assert self.db.DomainChange.select().count() == 6

        logWhoisQuery(
            self.db,
            logStatus(self.db, "foo"),
            domain,
            registrar_2,
            whois_server_2,
            creation_date_2,
            updated_date_2,
            expiration_date_2,
            name_servers_2,
            whois_status,
            emails,
            dnssec,
            name,
            org,
            address,
            city,
            state,
            zipcode,
            country,
        )
        assert self.db.DomainChange.select().count() == 7

        logWhoisQuery(
            self.db,
            logStatus(self.db, "foo"),
            domain,
            registrar_2,
            whois_server_2,
            creation_date_2,
            updated_date_2,
            expiration_date_2,
            name_servers_2,
            whois_status_2,
            emails,
            dnssec,
            name,
            org,
            address,
            city,
            state,
            zipcode,
            country,
        )
        assert self.db.DomainChange.select().count() == 8

        logWhoisQuery(
            self.db,
            logStatus(self.db, "foo"),
            domain,
            registrar_2,
            whois_server_2,
            creation_date_2,
            updated_date_2,
            expiration_date_2,
            name_servers_2,
            whois_status_2,
            emails_2,
            dnssec,
            name,
            org,
            address,
            city,
            state,
            zipcode,
            country,
        )
        assert self.db.DomainChange.select().count() == 9

        logWhoisQuery(
            self.db,
            logStatus(self.db, "foo"),
            domain,
            registrar_2,
            whois_server_2,
            creation_date_2,
            updated_date_2,
            expiration_date_2,
            name_servers_2,
            whois_status_2,
            emails_2,
            dnssec_2,
            name,
            org,
            address,
            city,
            state,
            zipcode,
            country,
        )
        assert self.db.DomainChange.select().count() == 10

        logWhoisQuery(
            self.db,
            logStatus(self.db, "foo"),
            domain,
            registrar_2,
            whois_server_2,
            creation_date_2,
            updated_date_2,
            expiration_date_2,
            name_servers_2,
            whois_status_2,
            emails_2,
            dnssec_2,
            name_2,
            org,
            address,
            city,
            state,
            zipcode,
            country,
        )
        assert self.db.DomainChange.select().count() == 11

        logWhoisQuery(
            self.db,
            logStatus(self.db, "foo"),
            domain,
            registrar_2,
            whois_server_2,
            creation_date_2,
            updated_date_2,
            expiration_date_2,
            name_servers_2,
            whois_status_2,
            emails_2,
            dnssec_2,
            name_2,
            org_2,
            address,
            city,
            state,
            zipcode,
            country,
        )
        assert self.db.DomainChange.select().count() == 12

        logWhoisQuery(
            self.db,
            logStatus(self.db, "foo"),
            domain,
            registrar_2,
            whois_server_2,
            creation_date_2,
            updated_date_2,
            expiration_date_2,
            name_servers_2,
            whois_status_2,
            emails_2,
            dnssec_2,
            name_2,
            org_2,
            address_2,
            city,
            state,
            zipcode,
            country,
        )
        assert self.db.DomainChange.select().count() == 13

        logWhoisQuery(
            self.db,
            logStatus(self.db, "foo"),
            domain,
            registrar_2,
            whois_server_2,
            creation_date_2,
            updated_date_2,
            expiration_date_2,
            name_servers_2,
            whois_status_2,
            emails_2,
            dnssec_2,
            name_2,
            org_2,
            address_2,
            city_2,
            state,
            zipcode,
            country,
        )
        assert self.db.DomainChange.select().count() == 14

        logWhoisQuery(
            self.db,
            logStatus(self.db, "foo"),
            domain,
            registrar_2,
            whois_server_2,
            creation_date_2,
            updated_date_2,
            expiration_date_2,
            name_servers_2,
            whois_status_2,
            emails_2,
            dnssec_2,
            name_2,
            org_2,
            address_2,
            city_2,
            state_2,
            zipcode,
            country,
        )
        assert self.db.DomainChange.select().count() == 15

        logWhoisQuery(
            self.db,
            logStatus(self.db, "foo"),
            domain,
            registrar_2,
            whois_server_2,
            creation_date_2,
            updated_date_2,
            expiration_date_2,
            name_servers_2,
            whois_status_2,
            emails_2,
            dnssec_2,
            name_2,
            org_2,
            address_2,
            city_2,
            state_2,
            zipcode_2,
            country,
        )
        assert self.db.DomainChange.select().count() == 16

        logWhoisQuery(
            self.db,
            logStatus(self.db, "foo"),
            domain,
            registrar_2,
            whois_server_2,
            creation_date_2,
            updated_date_2,
            expiration_date_2,
            name_servers_2,
            whois_status_2,
            emails_2,
            dnssec_2,
            name_2,
            org_2,
            address_2,
            city_2,
            state_2,
            zipcode_2,
            country_2,
        )
        assert self.db.DomainChange.select().count() == 17

    def test_logWhoisQuery_insertDifferentDomain(self):
        domain = "http://example.org"
        domain_2 = domain + "foo"
        registrar = "REGISTRAR"
        whois_server = "WHOISSERVER"
        creation_date = datetime.datetime.utcnow()
        updated_date = datetime.datetime.utcnow()
        expiration_date = datetime.datetime.utcnow()
        name_servers = "NAMESERVER"
        whois_status = "WHOISSTATUS"
        emails = "EMAIL"
        dnssec = "DNSSEC"
        name = "NAME"
        org = "ORG"
        address = "ADDRESS"
        city = "CITY"
        state = "STATE"
        zipcode = "ZIPCODE"
        country = "COUNTRY"
        status_id = logStatus(self.db, "foo")
        logWhoisQuery(
            self.db,
            status_id,
            domain,
            registrar,
            whois_server,
            creation_date,
            updated_date,
            expiration_date,
            name_servers,
            whois_status,
            emails,
            dnssec,
            name,
            org,
            address,
            city,
            state,
            zipcode,
            country,
        )
        logWhoisQuery(
            self.db,
            status_id,
            domain_2,
            registrar,
            whois_server,
            creation_date,
            updated_date,
            expiration_date,
            name_servers,
            whois_status,
            emails,
            dnssec,
            name,
            org,
            address,
            city,
            state,
            zipcode,
            country,
        )
        assert self.db.DomainChange.select().count() == 2

    ################################################
    # queryWhois
    ################################################
    def test_queryWhois_default(self):
        domain = "http://example.org"
        registrar = "REGISTRAR"
        whois_server = "WHOISSERVER"
        creation_date = datetime.datetime.utcnow()
        updated_date = datetime.datetime.utcnow()
        expiration_date = datetime.datetime.utcnow()
        name_servers = "NAMESERVER"
        whois_status = "WHOISSTATUS"
        emails = "EMAIL"
        dnssec = "DNSSEC"
        name = "NAME"
        org = "ORG"
        address = "ADDRESS"
        city = "CITY"
        state = "STATE"
        zipcode = "ZIPCODE"
        country = "COUNTRY"
        status_id = logStatus(self.db, "foo")

        mock_answer = MockAnswer(
            registrar,
            whois_server,
            creation_date,
            updated_date,
            expiration_date,
            name_servers,
            whois_status,
            emails,
            dnssec,
            name,
            org,
            address,
            city,
            state,
            zipcode,
            country,
        )
        with mock.patch("surykatka.domain.whois.whois") as mock_whois:
            mock_whois.return_value = mock_answer
            result = queryWhois(self.db, status_id, domain)

            assert mock_whois.call_count == 1
            mock_whois.assert_called_with(domain)

        assert self.db.DomainChange.select().count() == 1
        assert self.db.DomainChange.get().domain == domain
        assert self.db.DomainChange.get().registrar == registrar
        assert self.db.DomainChange.get().whois_server == whois_server
        assert self.db.DomainChange.get().creation_date == creation_date
        assert self.db.DomainChange.get().updated_date == updated_date
        assert self.db.DomainChange.get().expiration_date == expiration_date
        assert self.db.DomainChange.get().name_servers == name_servers
        assert self.db.DomainChange.get().whois_status == whois_status
        assert self.db.DomainChange.get().emails == emails
        assert self.db.DomainChange.get().dnssec == dnssec
        assert self.db.DomainChange.get().name == name
        assert self.db.DomainChange.get().org == org
        assert self.db.DomainChange.get().address == address
        assert self.db.DomainChange.get().city == city
        assert self.db.DomainChange.get().state == state
        assert self.db.DomainChange.get().zipcode == zipcode
        assert self.db.DomainChange.get().country == country
        assert self.db.DomainChange.get().status_id == status_id
        assert result == mock_answer

    ################################################
    # packDns
    ################################################
    def test_packDomain_oldLog(self):
        domain = "http://example.org"
        registrar = "REGISTRAR"
        registrar_2 = registrar + "foo"
        whois_server = "WHOISSERVER"
        creation_date = datetime.datetime(2000, 1, 1)
        updated_date = datetime.datetime(2000, 1, 3)
        expiration_date = datetime.datetime(2000, 1, 5)
        name_servers = "NAMESERVER"
        whois_status = "WHOISSTATUS"
        emails = "EMAIL"
        dnssec = "DNSSEC"
        name = "NAME"
        org = "ORG"
        address = "ADDRESS"
        city = "CITY"
        state = "STATE"
        zipcode = "ZIPCODE"
        country = "COUNTRY"

        logWhoisQuery(
            self.db,
            logStatus(self.db, "foo"),
            domain,
            registrar,
            whois_server,
            creation_date,
            updated_date,
            expiration_date,
            name_servers,
            whois_status,
            emails,
            dnssec,
            name,
            org,
            address,
            city,
            state,
            zipcode,
            country,
        )

        status_id_2 = logStatus(self.db, "foo")
        logWhoisQuery(
            self.db,
            status_id_2,
            domain,
            registrar_2,
            whois_server,
            creation_date,
            updated_date,
            expiration_date,
            name_servers,
            whois_status,
            emails,
            dnssec,
            name,
            org,
            address,
            city,
            state,
            zipcode,
            country,
        )
        assert self.db.DomainChange.select().count() == 2

        result = packDomain(self.db)
        assert self.db.DomainChange.select().count() == 1
        assert self.db.DomainChange.get().status_id == status_id_2
        assert result == None

    def test_packDns_keepDifferentUrl(self):
        domain = "http://example.org"
        domain_2 = domain + "foo"
        registrar = "REGISTRAR"
        whois_server = "WHOISSERVER"
        creation_date = datetime.datetime.utcnow()
        updated_date = datetime.datetime.utcnow()
        expiration_date = datetime.datetime.utcnow()
        name_servers = "NAMESERVER"
        whois_status = "WHOISSTATUS"
        emails = "EMAIL"
        dnssec = "DNSSEC"
        name = "NAME"
        org = "ORG"
        address = "ADDRESS"
        city = "CITY"
        state = "STATE"
        zipcode = "ZIPCODE"
        country = "COUNTRY"
        status_id = logStatus(self.db, "foo")
        logWhoisQuery(
            self.db,
            status_id,
            domain,
            registrar,
            whois_server,
            creation_date,
            updated_date,
            expiration_date,
            name_servers,
            whois_status,
            emails,
            dnssec,
            name,
            org,
            address,
            city,
            state,
            zipcode,
            country,
        )
        logWhoisQuery(
            self.db,
            status_id,
            domain_2,
            registrar,
            whois_server,
            creation_date,
            updated_date,
            expiration_date,
            name_servers,
            whois_status,
            emails,
            dnssec,
            name,
            org,
            address,
            city,
            state,
            zipcode,
            country,
        )
        assert self.db.DomainChange.select().count() == 2

        result = packDomain(self.db)
        assert self.db.DomainChange.select().count() == 2
        assert result == None


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(SurykatkaDomainTestCase))
    return suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")
