# Copyright (C) 2019  Nexedi SA and Contributors.
#                     Romain Courteaud <romain@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

import unittest
from surykatka.db import LogDB
import surykatka.network
from surykatka.network import logNetwork, isTcpPortOpen, packNetwork
from surykatka.status import logStatus
import mock
import peewee


class SurykatkaNetworkTestCase(unittest.TestCase):
    def setUp(self):
        self.db = LogDB(":memory:")
        self.db.createTables()

    ################################################
    # logNetwork
    ################################################
    def test_logNetwork_insertFirst(self):
        ip = "127.0.0.1"
        port = 1234
        transport = "foobar"
        state = "bar"
        status_id = logStatus(self.db, "foo")

        result = logNetwork(self.db, ip, transport, port, state, status_id)
        assert self.db.NetworkChange.select().count() == 1
        assert self.db.NetworkChange.get().ip == ip
        assert self.db.NetworkChange.get().port == port
        assert self.db.NetworkChange.get().transport == transport
        assert self.db.NetworkChange.get().state == state
        assert self.db.NetworkChange.get().status_id == status_id
        assert result == status_id

    def test_logNetwork_insertOnlyOnePerStatusIdIPPortTransport(self):
        ip = "127.0.0.1"
        port = 1234
        transport = "foobar"
        state = "bar"
        status_id = logStatus(self.db, "foo")

        logNetwork(self.db, ip, transport, port, state, status_id)
        try:
            logNetwork(self.db, ip, transport, port, state + ".", status_id)
        except peewee.IntegrityError:
            assert self.db.NetworkChange.select().count() == 1
            assert self.db.NetworkChange.get().status_id == status_id
        else:
            raise NotImplementedError("Expected IntegrityError")

    def test_logNetwork_skipIdenticalPreviousValues(self):
        ip = "127.0.0.1"
        port = 1234
        transport = "foobar"
        state = "bar"
        status_id = logStatus(self.db, "foo")
        result = logNetwork(self.db, ip, transport, port, state, status_id)

        status_id_2 = logStatus(self.db, "foo")
        result_2 = logNetwork(self.db, ip, transport, port, state, status_id_2)
        assert result_2 == result
        assert self.db.NetworkChange.select().count() == 1
        assert self.db.NetworkChange.get().ip == ip
        assert self.db.NetworkChange.get().port == port
        assert self.db.NetworkChange.get().transport == transport
        assert self.db.NetworkChange.get().state == state
        assert self.db.NetworkChange.get().status_id == status_id

    def test_logNetwork_insertWhenDifferentState(self):
        ip = "127.0.0.1"
        port = 1234
        transport = "foobar"
        state = "bar"
        status_id = logStatus(self.db, "foo")
        result = logNetwork(self.db, ip, transport, port, state, status_id)

        state_2 = state + "."
        status_id_2 = logStatus(self.db, "foo")
        result_2 = logNetwork(
            self.db, ip, transport, port, state_2, status_id_2
        )
        assert result_2 != result
        assert self.db.NetworkChange.select().count() == 2
        assert (
            self.db.NetworkChange.get(
                self.db.NetworkChange.status == status_id
            ).ip
            == ip
        )
        assert (
            self.db.NetworkChange.get(
                self.db.NetworkChange.status == status_id
            ).port
            == port
        )
        assert (
            self.db.NetworkChange.get(
                self.db.NetworkChange.status == status_id
            ).transport
            == transport
        )
        assert (
            self.db.NetworkChange.get(
                self.db.NetworkChange.status == status_id
            ).state
            == state
        )
        assert (
            self.db.NetworkChange.get(
                self.db.NetworkChange.status == status_id
            ).status_id
            == status_id
        )
        assert (
            self.db.NetworkChange.get(
                self.db.NetworkChange.status == status_id_2
            ).ip
            == ip
        )
        assert (
            self.db.NetworkChange.get(
                self.db.NetworkChange.status == status_id_2
            ).port
            == port
        )
        assert (
            self.db.NetworkChange.get(
                self.db.NetworkChange.status == status_id_2
            ).transport
            == transport
        )
        assert (
            self.db.NetworkChange.get(
                self.db.NetworkChange.status == status_id_2
            ).state
            == state_2
        )
        assert (
            self.db.NetworkChange.get(
                self.db.NetworkChange.status == status_id_2
            ).status_id
            == status_id_2
        )

    def test_logNetwork_insertDifferentKeys(self):
        ip = "127.0.0.1"
        ip_2 = ip + "2"
        port = 1234
        port_2 = port + 1
        transport = "foobar"
        transport_2 = transport + "."
        state = "bar"
        status_id = logStatus(self.db, "foo")

        logNetwork(self.db, ip, transport, port, state, status_id)
        logNetwork(self.db, ip_2, transport, port, state, status_id)
        logNetwork(self.db, ip, transport_2, port, state, status_id)
        logNetwork(self.db, ip, transport, port_2, state, status_id)
        logNetwork(self.db, ip_2, transport_2, port, state, status_id)
        logNetwork(self.db, ip_2, transport, port_2, state, status_id)
        logNetwork(self.db, ip, transport_2, port_2, state, status_id)
        logNetwork(self.db, ip_2, transport_2, port_2, state, status_id)

        assert self.db.NetworkChange.select().count() == 8

    ################################################
    # isTcpPortOpen
    ################################################
    def test_isTcpPortOpen_open(self):
        ip = "127.0.0.1"
        port = 1234
        status_id = logStatus(self.db, "foo")

        with mock.patch("surykatka.network.socket.socket") as mock_socket:
            result = isTcpPortOpen(self.db, ip, port, status_id)
            assert mock_socket.call_count == 1

            assert mock_socket.return_value.settimeout.call_count == 1
            mock_socket.return_value.settimeout.assert_called_with(2)

            assert mock_socket.return_value.connect.call_count == 1
            mock_socket.return_value.connect.assert_called_with((ip, port))

            assert mock_socket.return_value.close.call_count == 1

        assert self.db.NetworkChange.select().count() == 1
        assert self.db.NetworkChange.get().ip == ip
        assert self.db.NetworkChange.get().port == port
        assert self.db.NetworkChange.get().transport == "TCP"
        assert self.db.NetworkChange.get().state == "open"
        assert self.db.NetworkChange.get().status_id == status_id
        assert result == True

    def test_isTcpPortOpen_connectionRefused(self):
        ip = "127.0.0.1"
        port = 1234
        status_id = logStatus(self.db, "foo")

        with mock.patch("surykatka.network.socket.socket") as mock_socket:

            def sideEffect(*args, **kw):
                raise ConnectionRefusedError()

            mock_socket.return_value.connect.side_effect = sideEffect

            result = isTcpPortOpen(self.db, ip, port, status_id)
            assert mock_socket.call_count == 1

            assert mock_socket.return_value.settimeout.call_count == 1
            mock_socket.return_value.settimeout.assert_called_with(2)

            assert mock_socket.return_value.connect.call_count == 1
            mock_socket.return_value.connect.assert_called_with((ip, port))

            assert mock_socket.return_value.close.call_count == 1

        assert self.db.NetworkChange.select().count() == 1
        assert self.db.NetworkChange.get().ip == ip
        assert self.db.NetworkChange.get().port == port
        assert self.db.NetworkChange.get().transport == "TCP"
        assert self.db.NetworkChange.get().state == "closed"
        assert self.db.NetworkChange.get().status_id == status_id
        assert result == False

    def test_isTcpPortOpen_timeout(self):
        ip = "127.0.0.1"
        port = 1234
        status_id = logStatus(self.db, "foo")

        with mock.patch("surykatka.network.socket.socket") as mock_socket:

            def sideEffect(*args, **kw):
                raise surykatka.network.socket.timeout()

            mock_socket.return_value.connect.side_effect = sideEffect

            result = isTcpPortOpen(self.db, ip, port, status_id)
            assert mock_socket.call_count == 1

            assert mock_socket.return_value.settimeout.call_count == 1
            mock_socket.return_value.settimeout.assert_called_with(2)

            assert mock_socket.return_value.connect.call_count == 1
            mock_socket.return_value.connect.assert_called_with((ip, port))

            assert mock_socket.return_value.close.call_count == 1

        assert self.db.NetworkChange.select().count() == 1
        assert self.db.NetworkChange.get().ip == ip
        assert self.db.NetworkChange.get().port == port
        assert self.db.NetworkChange.get().transport == "TCP"
        assert self.db.NetworkChange.get().state == "filtered"
        assert self.db.NetworkChange.get().status_id == status_id
        assert result == False

    def test_isTcpPortOpen_timeouterror(self):
        ip = "127.0.0.1"
        port = 1234
        status_id = logStatus(self.db, "foo")

        with mock.patch("surykatka.network.socket.socket") as mock_socket:

            def sideEffect(*args, **kw):
                raise TimeoutError

            mock_socket.return_value.connect.side_effect = sideEffect

            result = isTcpPortOpen(self.db, ip, port, status_id)
            assert mock_socket.call_count == 1

            assert mock_socket.return_value.settimeout.call_count == 1
            mock_socket.return_value.settimeout.assert_called_with(2)

            assert mock_socket.return_value.connect.call_count == 1
            mock_socket.return_value.connect.assert_called_with((ip, port))

            assert mock_socket.return_value.close.call_count == 1

        assert self.db.NetworkChange.select().count() == 1
        assert self.db.NetworkChange.get().ip == ip
        assert self.db.NetworkChange.get().port == port
        assert self.db.NetworkChange.get().transport == "TCP"
        assert self.db.NetworkChange.get().state == "filtered"
        assert self.db.NetworkChange.get().status_id == status_id
        assert result == False

    def test_isTcpPortOpen_noRoute(self):
        ip = "127.0.0.1"
        port = 1234
        status_id = logStatus(self.db, "foo")

        with mock.patch("surykatka.network.socket.socket") as mock_socket:

            def sideEffect(*args, **kw):
                raise OSError(surykatka.network.errno.EHOSTUNREACH, "foo")

            mock_socket.return_value.connect.side_effect = sideEffect

            result = isTcpPortOpen(self.db, ip, port, status_id)
            assert mock_socket.call_count == 1

            assert mock_socket.return_value.settimeout.call_count == 1
            mock_socket.return_value.settimeout.assert_called_with(2)

            assert mock_socket.return_value.connect.call_count == 1
            mock_socket.return_value.connect.assert_called_with((ip, port))

            assert mock_socket.return_value.close.call_count == 1

        assert self.db.NetworkChange.select().count() == 1
        assert self.db.NetworkChange.get().ip == ip
        assert self.db.NetworkChange.get().port == port
        assert self.db.NetworkChange.get().transport == "TCP"
        assert self.db.NetworkChange.get().state == "filtered"
        assert self.db.NetworkChange.get().status_id == status_id
        assert result == False

    def test_isTcpPortOpen_noNetwork(self):
        ip = "127.0.0.1"
        port = 1234
        status_id = logStatus(self.db, "foo")

        with mock.patch("surykatka.network.socket.socket") as mock_socket:

            def sideEffect(*args, **kw):
                raise OSError(surykatka.network.errno.ENETUNREACH, "foo")

            mock_socket.return_value.connect.side_effect = sideEffect

            result = isTcpPortOpen(self.db, ip, port, status_id)
            assert mock_socket.call_count == 1

            assert mock_socket.return_value.settimeout.call_count == 1
            mock_socket.return_value.settimeout.assert_called_with(2)

            assert mock_socket.return_value.connect.call_count == 1
            mock_socket.return_value.connect.assert_called_with((ip, port))

            assert mock_socket.return_value.close.call_count == 1

        assert self.db.NetworkChange.select().count() == 1
        assert self.db.NetworkChange.get().ip == ip
        assert self.db.NetworkChange.get().port == port
        assert self.db.NetworkChange.get().transport == "TCP"
        assert self.db.NetworkChange.get().state == "unreachable"
        assert self.db.NetworkChange.get().status_id == status_id
        assert result == False

    def test_isTcpPortOpen_OSError(self):
        ip = "127.0.0.1"
        port = 1234
        status_id = logStatus(self.db, "foo")

        with mock.patch("surykatka.network.socket.socket") as mock_socket:

            def sideEffect(*args, **kw):
                raise OSError()

            mock_socket.return_value.connect.side_effect = sideEffect

            try:
                isTcpPortOpen(self.db, ip, port, status_id)
            except OSError:
                assert self.db.NetworkChange.select().count() == 0
            else:
                raise NotImplementedError("Expected OSError")

            assert mock_socket.call_count == 1

            assert mock_socket.return_value.settimeout.call_count == 1
            mock_socket.return_value.settimeout.assert_called_with(2)

            assert mock_socket.return_value.connect.call_count == 1
            mock_socket.return_value.connect.assert_called_with((ip, port))

            assert mock_socket.return_value.close.call_count == 1

    def test_isTcpPortOpen_Exception(self):
        ip = "127.0.0.1"
        port = 1234
        status_id = logStatus(self.db, "foo")

        with mock.patch("surykatka.network.socket.socket") as mock_socket:

            def sideEffect(*args, **kw):
                raise Exception()

            mock_socket.return_value.connect.side_effect = sideEffect

            try:
                isTcpPortOpen(self.db, ip, port, status_id)
            except Exception:
                assert self.db.NetworkChange.select().count() == 0
            else:
                raise NotImplementedError("Expected OSError")

            assert mock_socket.call_count == 1

            assert mock_socket.return_value.settimeout.call_count == 1
            mock_socket.return_value.settimeout.assert_called_with(2)

            assert mock_socket.return_value.connect.call_count == 1
            mock_socket.return_value.connect.assert_called_with((ip, port))

            assert mock_socket.return_value.close.call_count == 1

    ################################################
    # packNetwork
    ################################################
    def test_packNetwork_oldLog(self):
        ip = "127.0.0.1"
        port = 1234
        transport = "foobar"
        state = "bar"
        state_2 = "bar2"
        status_id = logStatus(self.db, "foo")
        status_id_2 = logStatus(self.db, "foo")

        logNetwork(self.db, ip, transport, port, state, status_id)
        logNetwork(self.db, ip, transport, port, state_2, status_id_2)

        result = packNetwork(self.db)
        assert self.db.NetworkChange.select().count() == 1
        assert self.db.NetworkChange.get().ip == ip
        assert self.db.NetworkChange.get().port == port
        assert self.db.NetworkChange.get().transport == transport
        assert self.db.NetworkChange.get().state == state_2
        assert self.db.NetworkChange.get().status_id == status_id_2
        assert result == None

    def test_packNetwork_keepDifferentUrl(self):
        ip = "127.0.0.1"
        ip_2 = ip + "2"
        port = 1234
        port_2 = port + 1
        transport = "foobar"
        transport_2 = transport + "."
        state = "bar"
        status_id = logStatus(self.db, "foo")
        logNetwork(self.db, ip, transport, port, state, status_id)
        status_id = logStatus(self.db, "foo")
        logNetwork(self.db, ip_2, transport, port, state, status_id)
        status_id = logStatus(self.db, "foo")
        logNetwork(self.db, ip, transport_2, port, state, status_id)
        status_id = logStatus(self.db, "foo")
        logNetwork(self.db, ip, transport, port_2, state, status_id)
        status_id = logStatus(self.db, "foo")
        logNetwork(self.db, ip_2, transport_2, port, state, status_id)
        status_id = logStatus(self.db, "foo")
        logNetwork(self.db, ip_2, transport, port_2, state, status_id)
        status_id = logStatus(self.db, "foo")
        logNetwork(self.db, ip, transport_2, port_2, state, status_id)
        status_id = logStatus(self.db, "foo")
        logNetwork(self.db, ip_2, transport_2, port_2, state, status_id)

        result = packNetwork(self.db)
        assert self.db.NetworkChange.select().count() == 8
        assert result == None


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(SurykatkaNetworkTestCase))
    return suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")
